import datetime
import logging
from typing import List

from metricflow.model.objects.materialization import Materialization
from metricflow.model.objects.user_configured_model import UserConfiguredModel
from metricflow.model.semantic_model import SemanticModel
from metricflow.model.semantics.data_source_container import PydanticDataSourceContainer
from metricflow.model.semantics.semantic_containers import DataSourceSemantics
from metricflow.model.validations.validator_helpers import (
    MaterializationContext,
    ModelValidationRule,
    ValidationError,
    ValidationIssueType,
    validate_safely,
    ValidationFutureError,
)
from metricflow.naming.linkable_spec_name import StructuredLinkableSpecName
from metricflow.query.query_parser import MetricFlowQueryParser
from metricflow.specs import TimeDimensionReference

logger = logging.getLogger(__name__)


class ValidMaterializationRule(ModelValidationRule):
    """Check that a materialization config is valid.

    * Metrics listed are valid.
    * Dimensions or identifiers are valid for the listed metrics.
    * Primary time dimension is listed in the dimensions section.
    """

    @staticmethod
    @validate_safely(whats_being_done="checking materialization is defined correctly")
    def _validate_materialization(
        materialization: Materialization,
        primary_time_dimensions_reference: TimeDimensionReference,
        mf_query_parser: MetricFlowQueryParser,
    ) -> List[ValidationIssueType]:
        issues: List[ValidationIssueType] = []

        context = MaterializationContext(
            file_name=materialization.metadata.file_slice.filename if materialization.metadata else None,
            line_number=materialization.metadata.file_slice.start_line_number if materialization.metadata else None,
            materialization_name=materialization.name,
        )

        if not materialization.dimensions:
            return [
                ValidationError(
                    context=context,
                    message=f"Materialization '{materialization.name}' does not have dimensions listed",
                )
            ]

        try:
            mf_query_parser.parse_and_validate_query(
                metric_names=materialization.metrics,
                group_by_names=materialization.dimensions,
            )
        # TODO: Using broad exception clause until the query validation returns a list of errors.
        except Exception as err:
            issues.append(
                ValidationFutureError(
                    context=context,
                    message=str(err),
                    error_date=datetime.date(2022, 5, 23),
                )
            )

        # Primary dimension checks
        mat_primary_time_dimension_names: List[str] = []
        for mat_dimension_or_identifier_name in materialization.dimensions:
            structured_spec = StructuredLinkableSpecName.from_name(mat_dimension_or_identifier_name)
            if structured_spec.element_name == primary_time_dimensions_reference.element_name:
                mat_primary_time_dimension_names.append(mat_dimension_or_identifier_name)

        if len(mat_primary_time_dimension_names) == 0:
            issues.append(
                ValidationError(
                    context=context,
                    message=f"Primary time dimension {primary_time_dimensions_reference.element_name} not listed"
                    f" as a dimension in materialization {materialization.name}",
                )
            )

        if len(mat_primary_time_dimension_names) > 1:
            issues.append(
                ValidationError(
                    context=context,
                    message=f"Multiple primary time dimensions {mat_primary_time_dimension_names} listed in "
                    f"materialization {materialization.name}",
                )
            )

        return issues

    @staticmethod
    @validate_safely(whats_being_done="running model validation ensuring materializations are valid")
    def validate_model(model: UserConfiguredModel) -> List[ValidationIssueType]:  # noqa: D
        """Check that all of the metrics and dimensions listed in a materialization are valid."""
        issues: List[ValidationIssueType] = []
        ds_semantics = DataSourceSemantics(
            model=model, configured_data_source_container=PydanticDataSourceContainer(model.data_sources)
        )
        primary_time_dimensions_reference = ds_semantics.primary_time_dimension_reference
        mf_query_parser = MetricFlowQueryParser(
            model=SemanticModel(user_configured_model=model),
            primary_time_dimension_reference=primary_time_dimensions_reference,
        )
        for materialization in model.materializations:
            issues += ValidMaterializationRule._validate_materialization(
                materialization=materialization,
                primary_time_dimensions_reference=primary_time_dimensions_reference,
                mf_query_parser=mf_query_parser,
            )
        return issues



from copy import deepcopy

import pytest
from _pytest.fixtures import FixtureRequest
from metricflow.engine.metricflow_engine import MetricFlowEngine
from metricflow.model.data_warehouse_model_validator import (
    DataWarehouseModelValidator,
    DataWarehouseTaskBuilder,
    DataWarehouseValidationTask,
)
from metricflow.model.model_transformer import ModelTransformer
from metricflow.model.objects.data_source import Mutability, MutabilityType
from metricflow.model.objects.elements.dimension import Dimension, DimensionType
from metricflow.model.objects.elements.measure import AggregationType, Measure
from metricflow.model.objects.user_configured_model import UserConfiguredModel
from metricflow.model.semantic_model import SemanticModel
from metricflow.model.validations.validator_helpers import ValidationIssueLevel
from metricflow.plan_conversion.column_resolver import DefaultColumnAssociationResolver
from metricflow.plan_conversion.time_spine import TimeSpineSource
from metricflow.protocols.sql_client import SqlClient
from metricflow.specs import DimensionReference, MeasureReference
from metricflow.test.fixtures.setup_fixtures import MetricFlowTestSessionState
from metricflow.test.model.validations.helpers import data_source_with_guaranteed_meta
from metricflow.test.plan_utils import assert_snapshot_text_equal, make_schema_replacement_function
from metricflow.test.test_utils import as_datetime
from metricflow.test.time.configurable_time_source import ConfigurableTimeSource


@pytest.fixture(scope="module")
def mf_engine(  # noqa: D
    data_warehouse_validation_model: UserConfiguredModel,
    sql_client: SqlClient,
    time_spine_source: TimeSpineSource,
    mf_test_session_state: MetricFlowTestSessionState,
) -> MetricFlowEngine:
    semantic_model = SemanticModel(data_warehouse_validation_model)
    return MetricFlowEngine(
        semantic_model=semantic_model,
        sql_client=sql_client,
        column_association_resolver=DefaultColumnAssociationResolver(semantic_model=semantic_model),
        time_source=ConfigurableTimeSource(as_datetime("2020-01-01")),
        time_spine_source=time_spine_source,
        system_schema=mf_test_session_state.mf_system_schema,
    )


def test_build_data_source_tasks(
    mf_test_session_state: MetricFlowTestSessionState, data_warehouse_validation_model: UserConfiguredModel
) -> None:  # noqa:D
    tasks = DataWarehouseTaskBuilder.gen_data_source_tasks(model=data_warehouse_validation_model)
    assert len(tasks) == len(data_warehouse_validation_model.data_sources)
    assert (
        tasks[0].query_string
        == f"SELECT (true) AS col0 FROM (SELECT * FROM {mf_test_session_state.mf_source_schema}.fct_animals) AS source0 "
        f"WHERE is_dog IS NOT NULL"
    )


def test_task_runner(sql_client: SqlClient, mf_engine: MetricFlowEngine) -> None:  # noqa: D
    dw_validator = DataWarehouseModelValidator(sql_client=sql_client, mf_engine=mf_engine)

    tasks = [
        DataWarehouseValidationTask(query_string="SELECT 'foo' AS foo", error_message="Could not select foo"),
        DataWarehouseValidationTask(query_string="SELECT 'bar' AS bar", error_message="Could not select bar"),
    ]

    issues = dw_validator.run_tasks(tasks=tasks)
    assert len(issues) == 0

    err_msg_bad = "Could not access table 'doesnt_exist' in data warehouse"
    bad_task = DataWarehouseValidationTask(
        query_string="SELECT (true) AS col1 FROM doesnt_exist", error_message=err_msg_bad
    )

    tasks.append(bad_task)
    issues = dw_validator.run_tasks(tasks=tasks)
    assert len(issues) == 1
    assert issues[0].level == ValidationIssueLevel.ERROR
    assert err_msg_bad in issues[0].message


def test_validate_data_sources(  # noqa: D
    data_warehouse_validation_model: UserConfiguredModel,
    create_data_warehouse_validation_model_tables: bool,
    sql_client: SqlClient,
    mf_engine: MetricFlowEngine,
) -> None:
    model = deepcopy(data_warehouse_validation_model)

    dw_validator = DataWarehouseModelValidator(sql_client=sql_client, mf_engine=mf_engine)

    issues = dw_validator.validate_data_sources(model)
    assert len(issues) == 0

    model.data_sources.append(
        data_source_with_guaranteed_meta(
            name="test_data_source2",
            sql_table="doesnt_exist",
            dimensions=[],
            mutability=Mutability(type=MutabilityType.IMMUTABLE),
        )
    )

    issues = dw_validator.validate_data_sources(model)
    assert len(issues) == 1
    assert issues[0].level == ValidationIssueLevel.ERROR
    assert "Unable to access data source `test_data_source2`" in issues[0].message


def test_build_dimension_tasks(  # noqa: D
    data_warehouse_validation_model: UserConfiguredModel, mf_test_session_state: MetricFlowTestSessionState
) -> None:
    tasks = DataWarehouseTaskBuilder.gen_dimension_tasks(model=data_warehouse_validation_model)
    assert len(tasks) == 1  # on data source query with all dimensions
    assert len(tasks[0].on_fail_subtasks) == 2  # a sub task for each dimension on the data source
    assert (
        f"SELECT (ds) AS col0, (is_dog) AS col1 FROM (SELECT * FROM {mf_test_session_state.mf_source_schema}.fct_animals) AS source0 WHERE is_dog IS NOT NULL"
        == tasks[0].query_string
    )
    assert (
        f"SELECT (ds) AS col0 FROM (SELECT * FROM {mf_test_session_state.mf_source_schema}.fct_animals) AS source0 WHERE is_dog IS NOT NULL"
        == tasks[0].on_fail_subtasks[0].query_string
    )
    assert (
        f"SELECT (is_dog) AS col0 FROM (SELECT * FROM {mf_test_session_state.mf_source_schema}.fct_animals) AS source0 WHERE is_dog IS NOT NULL"
        == tasks[0].on_fail_subtasks[1].query_string
    )


def test_validate_dimensions(  # noqa: D
    data_warehouse_validation_model: UserConfiguredModel, sql_client: SqlClient, mf_engine: MetricFlowEngine
) -> None:
    model = deepcopy(data_warehouse_validation_model)

    dw_validator = DataWarehouseModelValidator(sql_client=sql_client, mf_engine=mf_engine)

    issues = dw_validator.validate_dimensions(model)
    assert len(issues) == 0

    dimensions = list(model.data_sources[0].dimensions)
    dimensions.append(Dimension(name=DimensionReference(element_name="doesnt_exist"), type=DimensionType.CATEGORICAL))
    model.data_sources[0].dimensions = dimensions

    issues = dw_validator.validate_dimensions(model)
    # One isssue is created for the short circuit query failure, and another is
    # created for the sub task checking the specific dimension
    assert len(issues) == 2
    assert (
        f"Failed to query dimensions in data warehouse for data source `{model.data_sources[0].name}`"
        in issues[0].message
    )
    assert (
        f"Unable to query `doesnt_exist` in data warehouse for dimension `doesnt_exist` on data source `{model.data_sources[0].name}`"
        in issues[1].message
    )


def test_build_metric_tasks(  # noqa: D
    request: FixtureRequest,
    data_warehouse_validation_model: UserConfiguredModel,
    sql_client: SqlClient,
    mf_engine: MetricFlowEngine,
    mf_test_session_state: MetricFlowTestSessionState,
) -> None:
    tasks = DataWarehouseTaskBuilder.gen_metric_tasks(model=data_warehouse_validation_model, mf_engine=mf_engine)
    assert len(tasks) == 1
    assert_snapshot_text_equal(
        request=request,
        mf_test_session_state=mf_test_session_state,
        group_id="data_warehouse_validation_model",
        snapshot_id="query0",
        snapshot_text=tasks[0].query_string,
        snapshot_file_extension=".sql",
        incomparable_strings_replacement_function=make_schema_replacement_function(
            system_schema=mf_test_session_state.mf_system_schema, source_schema=mf_test_session_state.mf_source_schema
        ),
        additional_sub_directories_for_snapshots=(sql_client.__class__.__name__,),
    )


def test_validate_metrics(  # noqa: D
    data_warehouse_validation_model: UserConfiguredModel,
    sql_client: SqlClient,
    mf_engine: MetricFlowEngine,
    time_spine_source: TimeSpineSource,
    mf_test_session_state: MetricFlowTestSessionState,
) -> None:
    model = deepcopy(data_warehouse_validation_model)
    dw_validator = DataWarehouseModelValidator(sql_client=sql_client, mf_engine=mf_engine)

    issues = dw_validator.validate_metrics(model)
    assert len(issues) == 0

    # Update model to have a new measure which creates a new metric by proxy
    new_measures = list(model.data_sources[0].measures)
    new_measures.append(
        Measure(
            name=MeasureReference(element_name="count_cats"),
            agg=AggregationType.SUM,
            expr="is_cat",  # doesn't exist as column
            create_metric=True,
        )
    )
    model.data_sources[0].measures = new_measures
    model.metrics = []
    model = ModelTransformer.pre_validation_transform_model(model)
    model = ModelTransformer.post_validation_transform_model(model)

    # Get new metric flow engine which has the context of the updated model
    semantic_model = SemanticModel(model)
    new_engine = MetricFlowEngine(
        semantic_model=semantic_model,
        sql_client=sql_client,
        column_association_resolver=DefaultColumnAssociationResolver(semantic_model=semantic_model),
        time_source=ConfigurableTimeSource(as_datetime("2020-01-01")),
        time_spine_source=time_spine_source,
        system_schema=mf_test_session_state.mf_system_schema,
    )

    # Validate new metric created by proxy causes an issue (because the column used doesn't exist)
    dw_validator = DataWarehouseModelValidator(sql_client=sql_client, mf_engine=new_engine)
    issues = dw_validator.validate_metrics(model)
    assert len(issues) == 1
    assert issues[0].level == ValidationIssueLevel.ERROR
    assert "Unable to query metric `count_cats`" in issues[0].message



from metricflow.model.parsing.config_linter import ConfigLinter
from metricflow.model.validations.validator_helpers import ValidationIssueLevel


def test_lint_dir(config_linter_model_path: str) -> None:  # noqa: D
    issues = ConfigLinter().lint_dir(dir_path=config_linter_model_path)

    assert len(issues) == 1
    assert issues[0].level == ValidationIssueLevel.ERROR
    assert 'duplication of key "data_source"' in issues[0].as_readable_str()



from datetime import date

from metricflow.model.validations.validator_helpers import ValidationContext, ValidationFutureError


def test_validaiton_issue_to_json() -> None:  # noqa: D
    issue = ValidationFutureError(
        context=ValidationContext(file_name="foo.yaml", line_number=1337),
        message="A issue was found that will be an error",
        error_date=date(2022, 6, 13),
    )

    assert issue.json(), f"Failed to jsonify issue: {issue}"



from __future__ import annotations

from hashlib import sha1
from typing import List, Optional

from metricflow.errors.errors import ParsingException
from metricflow.model.objects.common import Metadata
from metricflow.model.objects.constraints.where import WhereClauseConstraint
from metricflow.model.objects.utils import ParseableObject, ParseableField, HashableBaseModel
from metricflow.object_utils import ExtendedEnum
from metricflow.specs import MeasureReference
from metricflow.time.time_granularity import TimeGranularity
from metricflow.time.time_granularity import string_to_time_granularity


class MetricType(ExtendedEnum):
    """Currently supported metric types"""

    MEASURE_PROXY = "measure_proxy"
    RATIO = "ratio"
    EXPR = "expr"
    CUMULATIVE = "cumulative"


class CumulativeMetricWindow(HashableBaseModel, ParseableField):
    """Describes the window of time the metric should be accumulated over. ie '1 day', '2 weeks', etc"""

    count: int
    granularity: TimeGranularity

    def to_string(self) -> str:  # noqa: D
        return f"{self.count} {self.granularity.value}"

    @staticmethod
    def parse(window: str) -> CumulativeMetricWindow:
        """Returns window values if parsing succeeds, None otherwise

        Output of the form: (<time unit count>, <time granularity>, <error message>) - error message is None if window is formatted properly
        """
        parts = window.split(" ")
        if len(parts) != 2:
            raise ParsingException(
                f"Invalid window ({window}) in cumulative metric. Should be of the form `<count> <granularity>`, ie `28 days`",
            )

        granularity = parts[1]
        # if we switched to python 3.9 this could just be `granularity = parts[0].removesuffix('s')
        if granularity.endswith("s"):
            # months -> month
            granularity = granularity[:-1]
        if granularity not in [item.value for item in TimeGranularity]:
            raise ParsingException(
                f"Invalid time granularity {granularity} in cumulative metric window string: ({window})",
            )

        count = parts[0]
        if not count.isdigit():
            raise ParsingException(f"Invalid count ({count}) in cumulative metric window string: ({window})")

        return CumulativeMetricWindow(
            count=int(count),
            granularity=string_to_time_granularity(granularity),
        )


class MetricTypeParams(HashableBaseModel, ParseableObject):
    """Type params add additional context to certain metric types (the context depends on the metric type)"""

    measure: Optional[MeasureReference]
    measures: Optional[List[MeasureReference]]
    numerator: Optional[MeasureReference]
    denominator: Optional[MeasureReference]
    expr: Optional[str]
    window: Optional[CumulativeMetricWindow]
    grain_to_date: Optional[TimeGranularity]


class Metric(HashableBaseModel, ParseableObject):
    """Describes a metric"""

    name: str
    description: Optional[str]
    type: MetricType
    type_params: MetricTypeParams
    constraint: Optional[WhereClauseConstraint]
    metadata: Optional[Metadata]

    @property
    def measure_names(self) -> List[MeasureReference]:  # noqa: D
        tp = self.type_params
        res = tp.measures or []
        if tp.measure:
            res.append(tp.measure)
        if tp.numerator:
            res.append(tp.numerator)
        if tp.denominator:
            res.append(tp.denominator)

        return res

    @property
    def definition_hash(self) -> str:  # noqa: D
        values: List[Optional[str]] = [self.name, self.type_params.expr or ""]
        if self.constraint:
            values.append(self.constraint.where)
            if self.constraint.linkable_names:
                values.extend(self.constraint.linkable_names)
        values.extend([m.element_name for m in self.measure_names])

        hash_builder = sha1()
        for s in values:
            if s is None:
                continue
            hash_builder.update(s.encode("utf-8"))
        return hash_builder.hexdigest()



from __future__ import annotations

from dataclasses import dataclass, field
from math import floor
from time import perf_counter
from typing import List, Optional

from metricflow.engine.metricflow_engine import MetricFlowEngine, MetricFlowExplainResult, MetricFlowQueryRequest
from metricflow.model.objects.data_source import DataSource
from metricflow.model.objects.user_configured_model import UserConfiguredModel
from metricflow.model.semantic_model import SemanticModel
from metricflow.model.validations.validator_helpers import (
    DataSourceContext,
    DimensionContext,
    MetricContext,
    ValidationContext,
    ValidationError,
    ValidationIssue,
    ValidationWarning,
)
from metricflow.protocols.sql_client import SqlClient
from metricflow.sql.sql_bind_parameters import SqlBindParameters


@dataclass
class DataWarehouseValidationTask:
    """Dataclass for defining a task to be used in the DataWarehouseModelValidator"""

    query_string: str
    error_message: str
    context: Optional[ValidationContext] = None
    query_params: SqlBindParameters = field(default_factory=lambda: SqlBindParameters())
    on_fail_subtasks: List[DataWarehouseValidationTask] = field(default_factory=lambda: [])


class DataWarehouseTaskBuilder:
    """Task builder for standard data warehouse validation tasks"""

    QUERY_TEMPLATE = "SELECT {columns_select} FROM {data_source} AS source{unique_id}"
    WHERE_TEMPLATE = " WHERE {where_filters}"
    PARTITION_COL_TEMPLATE = "{partition} IS NOT NULL"
    WRAPPED_COL_TEMPLATE = "({column}) AS col{unique_id}"

    @staticmethod
    def _wrap_col(col: str, id: int) -> str:  # noqa: D
        return DataWarehouseTaskBuilder.WRAPPED_COL_TEMPLATE.format(column=col, unique_id=id)

    @staticmethod
    def _where_clause_from_partitions(data_source: DataSource) -> str:
        """Takes the partitions for a data source and genrates a WHERE clause based on their definintions"""
        if not data_source.partitions:
            return ""

        where_stmts: List[str] = []
        for partition in data_source.partitions:
            element = f"({partition.expr})" if partition.expr else partition.name.element_name
            where_stmts.append(DataWarehouseTaskBuilder.PARTITION_COL_TEMPLATE.format(partition=element))
        return DataWarehouseTaskBuilder.WHERE_TEMPLATE.format(where_filters=" AND ".join(where_stmts))

    @staticmethod
    def _gen_query(data_source: DataSource, id: int, columns: List[str] = ["true"]) -> str:
        """Generates a basic sql query to select parts of them model definition

        Generates a basic sql query for verifying the model's definition with
        the data warehouse. For example if a data source with an sql_table
        value of 'table1' and no partition was passed in no columns list the
        query generated would be "SELECT (true) as col1 FROM table1 AS source0". If a the
        data source had a partition, say on the dimension 'dim1' then the query
        would be "SELECT (true) as col1 FROM table1 AS source0 WHERE dim1 IS NOT NULL".
        And if in addition columns ["dim2", "dim3"] were passed in then the
        query would be "SELECT (dim2) as col1, (dim3) as col2 FROM table1 AS source0 WHERE
        dim1 IS NOT NULL".

        Args:
            data_source: The data source to generate the query for
            id: A unique id used to alias the table reference
            columns: Column strings to select
        Returns:
            A string representing the query for the passed in arguments.
        """
        data_source_str = data_source.sql_table if data_source.sql_table else f"({data_source.sql_query})"

        wrapped_cols = [DataWarehouseTaskBuilder._wrap_col(col, index) for index, col in enumerate(columns)]
        columns_select = ", ".join(wrapped_cols)

        query = DataWarehouseTaskBuilder.QUERY_TEMPLATE.format(
            data_source=data_source_str, columns_select=columns_select, unique_id=id
        )
        query += DataWarehouseTaskBuilder._where_clause_from_partitions(data_source=data_source)
        return query

    @staticmethod
    def gen_data_source_tasks(model: UserConfiguredModel) -> List[DataWarehouseValidationTask]:
        """Generates a list of tasks for validating the data sources of the model"""
        tasks: List[DataWarehouseValidationTask] = []
        for index, data_source in enumerate(model.data_sources):
            tasks.append(
                DataWarehouseValidationTask(
                    query_string=DataWarehouseTaskBuilder._gen_query(data_source=data_source, id=index),
                    context=DataSourceContext(
                        file_name=data_source.metadata.file_slice.filename if data_source.metadata else None,
                        line_number=data_source.metadata.file_slice.start_line_number if data_source.metadata else None,
                        data_source_name=data_source.name,
                    ),
                    error_message=f"Unable to access data source `{data_source.name}` in data warehouse",
                )
            )
        return tasks

    @staticmethod
    def gen_dimension_tasks(model: UserConfiguredModel) -> List[DataWarehouseValidationTask]:
        """Generates a list of tasks for validating the dimensions of the model

        The high level tasks returned are "short cut" queries which try to
        query all the dimensions for a given data source. If that query fails,
        one or more of the dimensions is incorrectly specified. Thus if the
        query fails, there are subtasks which query the individual dimensions
        on the data source to identify which have issues.
        """

        tasks: List[DataWarehouseValidationTask] = []
        for index, data_source in enumerate(model.data_sources):
            # generate the subtasks
            data_source_tasks: List[DataWarehouseValidationTask] = []
            data_source_columns: List[str] = []
            for dimension in data_source.dimensions:
                dim_to_query = dimension.expr if dimension.expr is not None else dimension.name.element_name
                data_source_columns.append(dim_to_query)

                data_source_tasks.append(
                    DataWarehouseValidationTask(
                        query_string=DataWarehouseTaskBuilder._gen_query(
                            data_source=data_source, id=index, columns=[dim_to_query]
                        ),
                        context=DimensionContext(
                            file_name=data_source.metadata.file_slice.filename if data_source.metadata else None,
                            line_number=data_source.metadata.file_slice.start_line_number
                            if data_source.metadata
                            else None,
                            data_source_name=data_source.name,
                            dimension_name=dimension.name.element_name,
                        ),
                        error_message=f"Unable to query `{dim_to_query}` in data warehouse for dimension `{dimension.name.element_name}` on data source `{data_source.name}`.",
                    )
                )

            # generate the shortcut tasks with sub tasks
            tasks.append(
                DataWarehouseValidationTask(
                    query_string=DataWarehouseTaskBuilder._gen_query(
                        data_source=data_source,
                        id=index,
                        columns=data_source_columns,
                    ),
                    context=DataSourceContext(
                        file_name=data_source.metadata.file_slice.filename if data_source.metadata else None,
                        line_number=data_source.metadata.file_slice.start_line_number if data_source.metadata else None,
                        data_source_name=data_source.name,
                    ),
                    error_message=f"Failed to query dimensions in data warehouse for data source `{data_source.name}`",
                    on_fail_subtasks=data_source_tasks,
                )
            )
        return tasks

    @staticmethod
    def gen_metric_tasks(model: UserConfiguredModel, mf_engine: MetricFlowEngine) -> List[DataWarehouseValidationTask]:
        """Generates a list of tasks for validating the metrics of the model"""
        primary_time_dim = SemanticModel(
            user_configured_model=model
        ).data_source_semantics.primary_time_dimension_reference
        tasks: List[DataWarehouseValidationTask] = []
        for metric in model.metrics:
            mf_query = MetricFlowQueryRequest.create_with_random_request_id(
                metric_names=[metric.name], group_by_names=[primary_time_dim.element_name]
            )
            explain_result: MetricFlowExplainResult = mf_engine.explain(mf_request=mf_query)
            tasks.append(
                DataWarehouseValidationTask(
                    query_string=explain_result.rendered_sql.sql_query,
                    query_params=explain_result.rendered_sql.bind_parameters,
                    context=MetricContext(
                        file_name=metric.metadata.file_slice.filename if metric.metadata else None,
                        line_number=metric.metadata.file_slice.start_line_number if metric.metadata else None,
                        metric_name=metric.name,
                    ),
                    error_message=f"Unable to query metric `{metric.name}`.",
                )
            )
        return tasks


class DataWarehouseModelValidator:
    """A Validator for checking specific tasks for the model against the Data Warehouse

    Data Warehouse Validations are validations that are done against the data
    warehouse based on the model configured by the user. Their purpose is to
    ensure that queries generated by MetricFlow won't fail when you go to use
    them (assuming the model has passed these validations before use).
    """

    def __init__(self, sql_client: SqlClient, mf_engine: MetricFlowEngine) -> None:  # noqa: D
        self._sql_client = sql_client
        self._mf_engine = mf_engine

    def run_tasks(
        self, tasks: List[DataWarehouseValidationTask], timeout: Optional[int] = None
    ) -> List[ValidationIssue]:
        """Runs the list of tasks as queries agains the data warehouse, returning any found issues

        Args:
            tasks: A list of tasks to run against the data warehouse
            timeout: An optional timeout. Default is None. When the timeout is hit, function will return early.

        Returns:
            A list of validation issues discovered when running the passed in tasks against the data warehosue
        """

        # Used for keeping track if we go past the max time
        start_time = perf_counter()

        issues: List[ValidationIssue] = []
        # TODO: Asyncio implementation
        for index, task in enumerate(tasks):
            if timeout is not None and perf_counter() - start_time > timeout:
                issues.append(
                    ValidationWarning(
                        context=None,
                        message=f"Hit timeout before completing all tasks. Completed {index}/{len(tasks)} tasks.",
                    )
                )
                break
            try:
                self._sql_client.dry_run(stmt=task.query_string, sql_bind_parameters=task.query_params)
            except Exception as e:
                issues.append(
                    ValidationError(
                        context=task.context,
                        message=task.error_message + f"\nRecieved following error from data warehouse:\n{e}",
                    )
                )
                if task.on_fail_subtasks:
                    sub_task_timeout = floor(timeout - (perf_counter() - start_time)) if timeout else None
                    issues += self.run_tasks(tasks=task.on_fail_subtasks, timeout=sub_task_timeout)

        return issues

    def validate_data_sources(self, model: UserConfiguredModel, timeout: Optional[int] = None) -> List[ValidationIssue]:
        """Generates a list of tasks for validating the data sources of the model and then runs them

        Args:
            model: Model which to run data warehouse validations on
            timeout: An optional timeout. Default is None. When the timeout is hit, function will return early.

        Returns:
            A list of validation issues discovered when running the passed in tasks against the data warehosue
        """
        tasks = DataWarehouseTaskBuilder.gen_data_source_tasks(model=model)
        return self.run_tasks(tasks=tasks, timeout=timeout)

    def validate_dimensions(self, model: UserConfiguredModel, timeout: Optional[int] = None) -> List[ValidationIssue]:
        """Generates a list of tasks for validating the dimensions of the model and then runs them

        Args:
            model: Model which to run data warehouse validations on
            timeout: An optional timeout. Default is None. When the timeout is hit, function will return early.

        Returns:
            A list of validation issues. If there are no validation issues, an empty list is returned.
        """
        tasks = DataWarehouseTaskBuilder.gen_dimension_tasks(model=model)
        return self.run_tasks(tasks=tasks, timeout=timeout)

    def validate_metrics(self, model: UserConfiguredModel, timeout: Optional[int] = None) -> List[ValidationIssue]:
        """Generates a list of tasks for validating the metrics of the model and then runs them

        Args:
            model: Model which to run data warehouse validations on
            timeout: An optional timeout. Default is None. When the timeout is hit, function will return early.

        Returns:
            A list of validation issues. If there are no validation issues, an empty list is returned.
        """

        tasks = DataWarehouseTaskBuilder.gen_metric_tasks(model=model, mf_engine=self._mf_engine)
        return self.run_tasks(tasks=tasks, timeout=timeout)



import logging

import yaml
from jsonschema import exceptions

from metricflow.errors.errors import ParsingException
from metricflow.model.objects.common import YamlConfigFile
from metricflow.model.parsing.schemas_internal import (
    metric_validator,
    data_source_validator,
    materialization_validator,
)

VERSION_KEY = "mf_config_schema"
METRIC_TYPE = "metric"
DATA_SOURCE_TYPE = "data_source"
MATERIALIZATION_TYPE = "materialization"
DOCUMENT_TYPES = [METRIC_TYPE, DATA_SOURCE_TYPE, MATERIALIZATION_TYPE]

logger = logging.getLogger(__name__)


def validate_config_structure(config_yaml: YamlConfigFile) -> None:  # noqa: D
    """Validates config shape against jsonschema

    catches ValidationError and raise one exception at the end so
    we can get all the validation errors rather than just the first
    """
    errors = []
    for config_document in yaml.load_all(config_yaml.contents, Loader=yaml.SafeLoader):
        # The config document can be None if there is nothing but white space between two `---`
        # this isn't really an issue, so lets just swallow it
        if config_document is None:
            continue
        if not isinstance(config_document, dict):
            errors.append(
                str(
                    ParsingException(
                        f"Document is not a dict. Got `{type(config_document)}`: {config_document}",
                        config_filepath=config_yaml.filepath,
                    )
                )
            )
            continue
        for document_type in config_document.keys():
            try:
                if document_type == METRIC_TYPE:
                    metric_validator.validate(config_document[document_type])
                elif document_type == DATA_SOURCE_TYPE:
                    data_source_validator.validate(config_document[document_type])
                elif document_type == MATERIALIZATION_TYPE:
                    materialization_validator.validate(config_document[document_type])
                elif document_type == VERSION_KEY:
                    pass
                else:
                    raise ParsingException(
                        f"Invalid document type '{document_type}'. Valid document types are: {DOCUMENT_TYPES}",
                        config_filepath=config_yaml.filepath,
                    )
            except exceptions.ValidationError as e:
                errors.append(f"{e}")
                logger.exception(str(e))

    if len(errors) > 0:
        raise exceptions.ValidationError("\n".join(errors))



from __future__ import annotations

import functools
import traceback
from abc import ABC, abstractmethod
from dataclasses import dataclass
from datetime import date
from enum import Enum
from pydantic import BaseModel, Extra
from typing import Any, Callable, List, Optional, Tuple, Union

from metricflow.model.objects.elements.dimension import DimensionType
from metricflow.model.objects.user_configured_model import UserConfiguredModel

VALIDATE_SAFELY_ERROR_STR_TMPLT = ". Issue occurred in method `{method_name}` called with {arguments_str}"


class ValidationIssueLevel(Enum):
    """Categorize the issues found while validating a MQL model."""

    # Issue should be fixed, but model will still work in MQL
    WARNING = 0
    # Issue doesn't prevent model from working in MQL yet, but will eventually be an error
    FUTURE_ERROR = 1
    # Issue will prevent the model from working in MQL
    ERROR = 2
    # Issue is blocking and further validation should be stopped.
    FATAL = 3


class ModelObjectType(Enum):
    """Maps object types in the models to a readable string."""

    DATA_SOURCE = "data_source"
    MATERIALIZATION = "materialization"
    MEASURE = "measure"
    DIMENSION = "dimension"
    IDENTIFIER = "identifier"
    METRIC = "metric"


class ValidationContext(BaseModel):
    """The base context class for validation issues"""

    file_name: Optional[str]
    line_number: Optional[int]

    class Config:
        """Pydantic class configuration options"""

        extra = Extra.forbid

    def context_str(self) -> str:
        """Human readable stringified representation of the context"""

        context_string = ""

        if self.file_name:
            context_string += f"in file `{self.file_name}`"
            if self.line_number:
                context_string += f" on line #{self.line_number}"

        return context_string


class MaterializationContext(ValidationContext):
    """The context class for vaidation issues involving materializations"""

    materialization_name: str

    def context_str(self) -> str:
        """Human readable stringified representation of the context"""
        return f"with materialization `{self.materialization_name}` {ValidationContext.context_str(self)}"


class MetricContext(ValidationContext):
    """The context class for vaidation issues involving metrics"""

    metric_name: str

    def context_str(self) -> str:
        """Human readable stringified representation of the context"""
        return f"with metric `{self.metric_name}` {ValidationContext.context_str(self)}"


class DataSourceContext(ValidationContext):
    """The context class for vaidation issues involving data sources"""

    data_source_name: str

    def context_str(self) -> str:
        """Human readable stringified representation of the context"""
        return f"with data source `{self.data_source_name}` {ValidationContext.context_str(self)}"


class DimensionContext(DataSourceContext):
    """The context class for vaidation issues involving dimensions"""

    dimension_name: str

    def context_str(self) -> str:
        """Human readable stringified representation of the context"""
        return f"with dimension `{self.dimension_name}` in data source `{self.data_source_name}` {ValidationContext.context_str(self)}"


class IdentifierContext(DataSourceContext):
    """The context class for vaidation issues involving indentifiers"""

    identifier_name: str

    def context_str(self) -> str:
        """Human readable stringified representation of the context"""
        return f"with identifier `{self.identifier_name}` in data source `{self.data_source_name}` {ValidationContext.context_str(self)}"


class MeasureContext(DataSourceContext):
    """The context class for vaidation issues involving measures"""

    measure_name: str

    def context_str(self) -> str:
        """Human readable stringified representation of the context"""
        return f"with measure `{self.measure_name}` in data source `{self.data_source_name}` {ValidationContext.context_str(self)}"


class ValidationIssue(BaseModel):
    """An issue that was found while validating the MetricFlow model."""

    level: ValidationIssueLevel
    message: str
    context: Optional[ValidationContext]
    # Consider adding a enum here that categories the type of validation issue and standardize the error messages.

    def as_readable_str(self, with_level: bool = True) -> str:
        """Return a easily readable string that can be used to log the issue."""
        msg_base = f"{self.level.name} " if with_level else ""
        msg_base += self.context.context_str() if self.context else ""
        if msg_base:
            msg_base += " - "
        return msg_base + self.message


class ValidationWarning(ValidationIssue):
    """A warning that was found while validation the model."""

    level = ValidationIssueLevel.WARNING


class ValidationFutureError(ValidationIssue):
    """A future error that was found while validation the model."""

    error_date: date
    level = ValidationIssueLevel.FUTURE_ERROR

    def as_readable_str(self, with_level: bool = True) -> str:
        """Return a easily readable string that can be used to log the issue."""
        return (
            f"{ValidationIssue.as_readable_str(self, with_level=with_level)}"
            f"IMPORTANT: this error will break your model starting {self.error_date.strftime('%b %d, %Y')}. "
        )


class ValidationError(ValidationIssue):
    """An error that was found while validating the model."""

    level = ValidationIssueLevel.ERROR


class ValidationFatal(ValidationIssue):
    """A fatal issue that was found while validation the model."""

    level = ValidationIssueLevel.FATAL


ValidationIssueType = Union[ValidationIssue, ValidationWarning, ValidationFutureError, ValidationError, ValidationFatal]


def generate_exception_issue(
    what_was_being_done: str,
    e: Exception,
    context: Optional[ValidationContext] = None,
) -> ValidationIssue:
    """Generates a validation issue for exceptions"""
    return ValidationError(
        context=context,
        message=f"An error occured while {what_was_being_done} - {''.join(traceback.format_exception(etype=type(e), value=e, tb=e.__traceback__))}",
    )


def _func_args_to_string(*args: Any, **kwargs: Any) -> str:  # type: ignore
    return f"positional args: {args}, key word args: {kwargs}"


def validate_safely(whats_being_done: str) -> Callable:
    """Decorator to safely run validation checks"""

    def decorator_check_element_safely(func: Callable) -> Callable:  # noqa
        @functools.wraps(func)
        def wrapper(*args: Any, **kwargs: Any) -> List[ValidationIssueType]:  # type: ignore
            """Safely run a check on model elements"""
            issues: List[ValidationIssueType]
            try:
                issues = func(*args, **kwargs)
            except Exception as e:
                arguments_str = _func_args_to_string(*args, **kwargs)
                issues = [
                    generate_exception_issue(
                        what_was_being_done=whats_being_done
                        + VALIDATE_SAFELY_ERROR_STR_TMPLT.format(
                            method_name=func.__name__, arguments_str=arguments_str
                        ),
                        e=e,
                    )
                ]
            return issues

        return wrapper

    return decorator_check_element_safely


@dataclass(frozen=True)
class DimensionInvariants:
    """Helper object to ensure consistent dimension attributes across data sources.

    All dimensions with a given name in all data sources should have attributes matching these values.
    """

    type: DimensionType
    is_partition: bool


class ModelValidationRule(ABC):
    """Encapsulates logic for checking the values of objects in a model."""

    @staticmethod
    @abstractmethod
    def validate_model(model: UserConfiguredModel) -> List[ValidationIssueType]:
        """Check the given model and return a list of validation issues"""
        pass


class ModelValidationException(Exception):
    """Exception raised when validation of a model fails."""

    def __init__(self, issues: Tuple[ValidationIssueType, ...]) -> None:  # noqa: D
        issues_str = "\n".join([x.as_readable_str() for x in issues])
        super().__init__(f"Error validating model. Issues:\n{issues_str}")



from __future__ import annotations

import re
from typing import Optional

from metricflow.errors.errors import ParsingException
from metricflow.model.objects.utils import HashableBaseModel
from metricflow.object_utils import ExtendedEnum
from metricflow.specs import LinkableElementReference


class Version(HashableBaseModel):  # noqa: D
    major: int
    minor: int

    _VERSION_REGEX = re.compile(r"^v[0-9]+\.[0-9]+$")

    @staticmethod
    def parse(version: str) -> Version:  # noqa: D

        if not Version._VERSION_REGEX.match(version):
            raise ParsingException(
                f"Expected a version of the form 'v<MAJOR_VERSION>.<MINOR_VERSION>' but got '{version}'."
            )
        if version[0] == "v":
            version = version[1:]

        parts = version.split(".")
        assert len(parts) == 2

        return Version(major=int(parts[0]), minor=int(parts[1]))

    def __str__(self) -> str:  # noqa: D
        return f"{self.__class__.__name__}(major={self.major}, minor={self.minor})"


class Element:  # noqa: D
    name: LinkableElementReference
    expr: Optional[str]
    type: ExtendedEnum

    @property
    def is_primary_time(self) -> bool:  # noqa: D
        raise NotImplementedError(
            f"Subclasses of Element must implement `is_primary_time`. This object is of type {type(self)}"
        )


class SourceFile(HashableBaseModel):  # noqa: D
    path: str
    contents: str


class Commit(HashableBaseModel):  # noqa: D
    commit: str
    timestamp: int


class FileSlice(HashableBaseModel):  # noqa: D
    filename: str
    content: str
    start_line_number: int
    end_line_number: int


class YamlConfigFile(HashableBaseModel):  # noqa: D
    filepath: str
    contents: str
    url: Optional[str]


class SourceContext(HashableBaseModel):  # noqa: D
    definition_hash: str


class Metadata(HashableBaseModel):  # noqa: D
    repo_file_path: str
    file_slice: FileSlice



import copy
import logging
from typing import List, Sequence

from metricflow.model.objects.user_configured_model import UserConfiguredModel
from metricflow.model.parsing.dir_to_model import ModelBuildResult
from metricflow.model.validations.data_sources import (
    DataSourceMeasuresUniqueRule,
    DataSourceTimeDimensionWarningsRule,
)
from metricflow.model.validations.dimension_const import DimensionConsistencyRule
from metricflow.model.validations.element_const import ElementConsistencyRule
from metricflow.model.validations.identifiers import (
    IdentifierConfigRule,
    OnePrimaryIdentifierPerDataSourceRule,
    IdentifierConsistencyRule,
)
from metricflow.model.validations.materializations import ValidMaterializationRule
from metricflow.model.validations.metrics import MetricMeasuresRule, CumulativeMetricRule
from metricflow.model.validations.non_empty import NonEmptyRule
from metricflow.model.validations.unique_valid_name import UniqueAndValidNameRule
from metricflow.model.validations.validator_helpers import (
    ValidationIssueType,
    ModelValidationRule,
    ValidationIssueLevel,
    ModelValidationException,
)

logger = logging.getLogger(__name__)


class ModelValidator:
    """A Validator that acts on UserConfiguredModel"""

    DEFAULT_RULES = (
        DataSourceMeasuresUniqueRule(),
        DataSourceTimeDimensionWarningsRule(),
        DimensionConsistencyRule(),
        ElementConsistencyRule(),
        IdentifierConfigRule(),
        IdentifierConsistencyRule(),
        OnePrimaryIdentifierPerDataSourceRule(),
        MetricMeasuresRule(),
        CumulativeMetricRule(),
        NonEmptyRule(),
        UniqueAndValidNameRule(),
        ValidMaterializationRule(),
    )

    def __init__(self, rules: Sequence[ModelValidationRule] = DEFAULT_RULES) -> None:
        """Constructor.

        Args:
            rules: List of validation rules to run. Defaults to DEFAULT_RULES
        """

        # Raises an error if 'rules' is an empty sequence or None
        if not rules:
            raise ValueError("ModelValidator 'rules' must be a sequence with at least one ModelValidationRule.")

        self._rules = rules

    def validate_model(self, model: UserConfiguredModel) -> ModelBuildResult:
        """Validate a model according to configured rules."""
        model_copy = copy.deepcopy(model)

        issues: List[ValidationIssueType] = []
        for validation_rule in self._rules:
            issues.extend(validation_rule.validate_model(model_copy))
            # If there are any fatal errors, stop the validation process.
            if any([x.level == ValidationIssueLevel.FATAL for x in issues]):
                return ModelBuildResult(model=model_copy, issues=tuple(issues))

        # If there are any errors, don't run any transforms and return the issues found.
        if any([x.level == ValidationIssueLevel.ERROR for x in issues]):
            return ModelBuildResult(model=model_copy, issues=tuple(issues))

        return ModelBuildResult(model=model_copy, issues=tuple(issues))

    def checked_validations(self, model: UserConfiguredModel) -> UserConfiguredModel:  # chTODO: remember checked_build
        """Similar to validate(), but throws an exception if validation fails."""
        model_copy = copy.deepcopy(model)
        build_result = self.validate_model(model_copy)
        if build_result.issues is not None:
            if any(
                [
                    x.level == ValidationIssueLevel.WARNING or x.level == ValidationIssueLevel.FUTURE_ERROR
                    for x in build_result.issues
                ]
            ):
                issues_str = "\n".join([x.as_readable_str() for x in build_result.issues])
                logger.warning(f"Found some validation warnings in the model:\n{issues_str}")
            if any(
                [
                    x.level == ValidationIssueLevel.ERROR or x.level == ValidationIssueLevel.FATAL
                    for x in build_result.issues
                ]
            ):
                raise ModelValidationException(issues=build_result.issues)
        return model
